import inspect
import json
import logging
import uuid

_LOGGER = logging.getLogger(__name__)


class JSONRPCError(Exception):
    """Root exception for all errors related to this library"""


class TransportError(JSONRPCError):
    """An error occurred while performing a connection to the server"""

    def __init__(self, exception_text, message=None, *args):
        """Create the transport error for the attempted message."""
        if message:
            super(TransportError, self).__init__(
                '%s: %s' % (message.transport_error_text, exception_text),
                *args)
        else:
            super(TransportError, self).__init__(exception_text, *args)


class ProtocolError(JSONRPCError):
    """An error occurred while dealing with the JSON-RPC protocol"""


class Server(object):
    """A connection to a JSON-RPC server"""

    def __init__(self):
        """Initialize the json-rpc server object."""
        self._server_request_handlers = {}

    def send_message(self, message):
        """Issue the request to the server and return the method result.

        This method must be implemented by the child class.
        """
        raise NotImplementedError()

    def receive_request(self, request):
        """Called when a request is received from the server."""
        result = None
        error = None
        args, kwargs = request.get_args()

        if request.method in self._server_request_handlers:
            try:
                handler = self._server_request_handlers[request.method]
                if inspect.iscoroutinefunction(handler):
                    raise TypeError(
                        "Async handlers are not supported in "
                        "synchronous sever implementations")
                else:
                    result = handler(*args, **kwargs)
            except Exception as exc:
                _LOGGER.error(exc, exc_info=exc)
                error = {
                    'code': -32000,
                    'message': 'Server Error: %s' % exc,
                }
        else:
            error = {
                'code': -32601,
                'message': 'Method not found',
            }

        if request.msg_id is not None:
            return Response(request, result, error)
        else:
            return None

    async def async_receive_request(self, request):
        """Called when a request is received from the server.

        If the implementation calls async_receive_request instead of
        receive_request, asynchronous request handlers are also supported
        """
        result = None
        error = None
        args, kwargs = request.get_args()

        if request.method in self._server_request_handlers:
            try:
                handler = self._server_request_handlers[request.method]
                if inspect.iscoroutinefunction(handler):
                    result = await handler(*args, **kwargs)
                else:
                    result = handler(*args, **kwargs)
            except Exception as exc:
                _LOGGER.error(exc, exc_info=exc)
                error = {
                    'code': -32000,
                    'message': 'Server Error: %s' % exc,
                }
        else:
            error = {
                'code': -32601,
                'message': 'Method not found',
            }

        if request.msg_id is not None:
            return Response(request, result, error)
        else:
            return None

    def __getattr__(self, method_name):
        if method_name.startswith("_"):  # prevent calls for private methods
            raise AttributeError("invalid attribute '%s'" % method_name)
        return Method(self.__request, self.__register, method_name)

    def __setattr__(self, method_name, callback):
        if method_name.startswith("_"):  # prevent calls for private methods
            return super(Server, self).__setattr__(method_name, callback)
        return self.__register(method_name, callback)

    def __request(self, method_name, args=None, kwargs=None):
        """Perform the actual RPC call.

        If _notification=True, don't wait for a response
        """
        if kwargs.pop('_notification', False):
            msg_id = None
        else:
            msg_id = str(uuid.uuid4())

        if args and kwargs:
            raise ProtocolError(
                'JSON-RPC spec forbids mixing arguments and keyword arguments')

        return self.send_message(Request(method_name, args or kwargs, msg_id))

    def __register(self, method_name, callback):
        """Register a callback for if the server sends this request."""
        self._server_request_handlers[method_name] = callback


class Message(object):
    """Message to be sent to the jsonrpc server."""

    @property
    def response_id(self):
        return None

    def serialize(self):
        """Generate the raw JSON message to be sent to the server"""
        raise NotImplementedError()

    def parse_response(self, response):
        """Parse the response from the server and return the result."""
        raise NotImplementedError()

    @property
    def transport_error_text(self):
        """Exception text for a transport error."""
        raise NotImplementedError()

    def __str__(self):
        return self.serialize()


class Request(Message):
    """Request a method call on the server."""

    def __init__(self, method=None, params=None, msg_id=None):
        self.method = method
        self.params = params
        self.msg_id = msg_id

    @staticmethod
    def parse(data):
        """Generate a request object by parsing the json data."""
        if 'method' not in data:
            raise ProtocolError('Request from server does not contain method')
        method = data.get('method')
        params = data.get('params')
        msg_id = data.get('id')
        if (
                not isinstance(params, list)
                and not isinstance(params, dict)
                and params is not None):
            raise ProtocolError(
                'Parameters must either be a positional list or named dict.')
        return Request(method, params, msg_id)

    @property
    def response_id(self):
        return self.msg_id

    def serialize(self):
        """Generate the raw JSON message to be sent to the server"""
        data = {'jsonrpc': '2.0', 'method': self.method}
        if self.params:
            data['params'] = self.params
        if self.msg_id is not None:
            data['id'] = self.msg_id
        return json.dumps(data)

    def parse_response(self, data):
        """Parse the response from the server and return the result."""
        if self.msg_id is None:
            # Don't parse results for notification requests
            return None

        if not isinstance(data, dict):
            raise ProtocolError('Response is not a dictionary')
        if data.get('error') is not None:
            code = data['error'].get('code', '')
            message = data['error'].get('message', '')
            raise ProtocolError(code, message, data)
        elif 'result' not in data:
            raise ProtocolError('Response without a result field')
        else:
            return data['result']

    @property
    def transport_error_text(self):
        """Exception text for a transport error."""
        return 'Error calling method %r' % self.method

    def get_args(self):
        """Transform the request parameters into args/kwargs"""
        args = []
        kwargs = {}
        if isinstance(self.params, list):
            args = self.params
        elif isinstance(self.params, dict):
            kwargs = self.params
        elif self.params is not None:
            raise ProtocolError(
                'Parameters must either be a positional list or named dict.')
        return args, kwargs


class Response(Message):
    """Respond to a method call from the server."""

    def __init__(self, request, result=None, error=None):
        self.request = request
        self.result = result
        self.error = error

    def serialize(self):
        """Generate the raw JSON message to be sent to the server"""
        data = {'jsonrpc': '2.0', 'id': self.request.msg_id}
        if self.error is not None:
            data['error'] = self.error
        else:
            data['result'] = self.result
        return json.dumps(data)

    def parse_response(self, response):
        """Parse the response from the server and return the result."""
        # Don't parse results from response messages
        return None

    @property
    def transport_error_text(self):
        """Exception text for a transport error."""
        return 'Error responding to server method %r' % self.request.method


class Method(object):
    """Map the methods called on the server to json-rpc methods."""

    def __init__(self, request_method, register_method, method_name):
        self.__request_method = request_method
        self.__register_method = register_method
        self.__method_name = method_name

    def __getattr__(self, method_name):
        if method_name.startswith("_"):  # prevent calls for private methods
            raise AttributeError("invalid attribute '%s'" % method_name)
        return Method(self.__request_method, self.__register_method,
                      "%s.%s" % (self.__method_name, method_name))

    def __call__(self, *args, **kwargs):
        return self.__request_method(self.__method_name, args, kwargs)

    def __setattr__(self, method_name, callback):
        if method_name.startswith("_"):  # prevent calls for private methods
            return super(Method, self).__setattr__(method_name, callback)
        return self.__register_method(
            "%s.%s" % (self.__method_name, method_name), callback)



import json

import pytest
from unittest import mock

import jsonrpc_base
from jsonrpc_base import Server, ProtocolError, TransportError

pytestmark = pytest.mark.asyncio


class MockTransportError(ValueError):
    """Test exception representing a transport library error."""


class MockServer(Server):

    def __init__(self, url):
        super(MockServer, self).__init__()
        self._url = url
        self._handler = None

    def send_message(self, message):
        """Issue the request to the server and return the method result"""
        try:
            if isinstance(message, jsonrpc_base.Request):
                data = jsonrpc_base.Request.parse(
                    json.loads(message.serialize()))
            else:
                data = message.serialize()
            response = json.loads(json.dumps(self._handler(data)))
        except Exception as requests_exception:
            raise TransportError(
                'Transport Error', message, requests_exception)

        return message.parse_response(response)


def assertSameJSON(json1, json2):
    """Tells whether two json strings, once decoded, are the same dictionary"""
    assert json.loads(json1) == json.loads(json2)


@pytest.fixture
def server():
    """Get the mock server object"""
    return MockServer('http://mock/xmlrpc')


def test_dumps(server):
    # test no args
    assertSameJSON(
        '''{"jsonrpc": "2.0", "method": "my_method_name", "id": 1}''',
        jsonrpc_base.Request(
            'my_method_name', params=None, msg_id=1).serialize()
    )
    # test zero message ID
    assertSameJSON(
        '''{"jsonrpc": "2.0", "method": "my_method_name", "id": 0}''',
        jsonrpc_base.Request(
            'my_method_name', params=None, msg_id=0).serialize()
    )
    # test empty args dict
    assertSameJSON(
        '''{"jsonrpc": "2.0", "method": "my_method_name", "id": 1}''',
        jsonrpc_base.Request(
            'my_method_name', params={}, msg_id=1).serialize()
    )
    # test empty args array
    assertSameJSON(
        '''{"jsonrpc": "2.0", "method": "my_method_name", "id": 1}''',
        jsonrpc_base.Request(
            'my_method_name', params=[], msg_id=1).serialize()
    )
    # test keyword args
    assertSameJSON(
        '''{"params": {"foo": "bar"}, "jsonrpc": "2.0", "method":
        "my_method_name", "id": 1}''',
        jsonrpc_base.Request(
            'my_method_name', params={'foo': 'bar'}, msg_id=1).serialize()
    )
    # test positional args
    assertSameJSON(
        '''{"params": ["foo", "bar"], "jsonrpc": "2.0", "method":
        "my_method_name", "id": 1}''',
        jsonrpc_base.Request(
            'my_method_name', params=('foo', 'bar'), msg_id=1).serialize()
    )
    # test notification
    assertSameJSON(
        '''{"params": ["foo", "bar"], "jsonrpc": "2.0", "method":
        "my_method_name"}''',
        jsonrpc_base.Request(
            'my_method_name', params=('foo', 'bar'), msg_id=None).serialize()
    )


def test_parse_result(server):
    request = jsonrpc_base.Request('my_message', msg_id=1)
    with pytest.raises(ProtocolError, match='Response is not a dictionary'):
        request.parse_response([])
    with pytest.raises(ProtocolError, match='Response without a result field'):
        request.parse_response({})
    with pytest.raises(ProtocolError) as protoerror:
        body = {
            "jsonrpc": "2.0", "error": {
                "code": -32601, "message": "Method not found"
            }, "id": "1"
        }
        request.parse_response(body)
    assert protoerror.value.args[0] == -32601
    assert protoerror.value.args[1] == 'Method not found'


def test_send_message(server):
    empty_server = Server()
    with pytest.raises(NotImplementedError):
        empty_server.send_message(jsonrpc_base.Request('my_method', msg_id=1))

    # catch non-json responses
    with pytest.raises(TransportError) as transport_error:
        def handler(message):
            raise MockTransportError("Transport Error")

        server._handler = handler
        server.send_message(jsonrpc_base.Request('my_method', msg_id=1))

    assert transport_error.value.args[0] == (
        "Error calling method 'my_method': Transport Error")
    assert isinstance(transport_error.value.args[1], MockTransportError)

    # a notification
    def handler(message):
        return 'we dont care about this'
    server._handler = handler
    server.send_message(jsonrpc_base.Request('my_notification', msg_id=None))


def test_exception_passthrough(server):
    with pytest.raises(TransportError) as transport_error:
        def handler(message):
            raise MockTransportError("Transport Error")
        server._handler = handler
        server.foo()
    assert transport_error.value.args[0] == (
        "Error calling method 'foo': Transport Error")
    assert isinstance(transport_error.value.args[1], MockTransportError)


def test_transport_error_constructor(server):
    with pytest.raises(TransportError, match='Test Message'):
        raise TransportError('Test Message')


def test_forbid_private_methods(server):
    """Test that we can't call private class methods"""
    with pytest.raises(AttributeError):
        server._foo()

    # nested private method call
    with pytest.raises(AttributeError):
        server.foo.bar._baz()


def test_method_call(server):
    """mixing *args and **kwargs is forbidden by the spec"""
    with pytest.raises(
            ProtocolError,
            match='JSON-RPC spec forbids mixing arguments and keyword '
            'arguments'):
        server.testmethod(1, 2, a=1, b=2)


def test_method_nesting(server):
    """Test that we correctly nest namespaces"""
    def handler(message):
        return {
            "jsonrpc": "2.0",
            "result": True if message.params[0] == message.method else False,
            "id": 1,
        }
    server._handler = handler

    assert server.nest.testmethod("nest.testmethod")
    assert server.nest.testmethod.some.other.method(
        "nest.testmethod.some.other.method")


def test_calls(server):
    # rpc call with positional parameters:
    def handler1(message):
        assert message.msg_id == "abcd-1234"
        assert message.params == [42, 23]
        return {
            "jsonrpc": "2.0",
            "result": 19,
            "id": 1,
        }

    server._handler = handler1
    with mock.patch("uuid.uuid4", return_value="abcd-1234"):
        assert server.subtract(42, 23) == 19

    # rpc call with named parameters
    def handler2(message):
        assert message.params == {'y': 23, 'x': 42}
        return {
            "jsonrpc": "2.0",
            "result": 19,
            "id": 1,
        }

    server._handler = handler2
    assert server.subtract(x=42, y=23) == 19

    # rpc call with a mapping type
    def handler3(message):
        assert message.params == [{'foo': 'bar'}]
        return {
            "jsonrpc": "2.0",
            "result": None,
            "id": 1,
        }

    server._handler = handler3
    server.foobar({'foo': 'bar'})

    # rpc call with direct dict params
    def handler3(message):
        assert message.params == {'foo': 'bar'}
        return {
            "jsonrpc": "2.0",
            "result": None,
            "id": 1,
        }

    server._handler = handler3
    server.foobar(**{'foo': 'bar'})


def test_notification(server):
    # Verify that we ignore the server response
    def handler(message):
        return {
            "jsonrpc": "2.0",
            "result": 19,
            "id": 3,
        }
    server._handler = handler
    assert server.subtract(42, 23, _notification=True) is None


def test_receive_server_requests(server):
    def event_handler(*args, **kwargs):
        return args, kwargs
    server.on_server_event = event_handler
    server.namespace.on_server_event = event_handler

    response = server.receive_request(jsonrpc_base.Request(
        'on_server_event', msg_id=1))
    args, kwargs = response.result
    assert len(args) == 0
    assert len(kwargs) == 0

    # Test with a zero message ID
    response = server.receive_request(jsonrpc_base.Request(
        'on_server_event', msg_id=0))
    args, kwargs = response.result
    assert len(args) == 0
    assert len(kwargs) == 0

    response = server.receive_request(jsonrpc_base.Request(
        'namespace.on_server_event', msg_id=1))
    args, kwargs = response.result
    assert len(args) == 0
    assert len(kwargs) == 0

    response = server.receive_request(jsonrpc_base.Request(
        'on_server_event', params=['foo', 'bar'], msg_id=1))
    args, kwargs = response.result
    assert args == ('foo', 'bar')
    assert len(kwargs) == 0

    response = server.receive_request(jsonrpc_base.Request(
        'on_server_event', params={'foo': 'bar'}, msg_id=1))
    args, kwargs = response.result
    assert len(args) == 0
    assert kwargs == {'foo': 'bar'}

    with pytest.raises(ProtocolError):
        response = server.receive_request(jsonrpc_base.Request(
            'on_server_event', params="string_params", msg_id=1))

    response = server.receive_request(jsonrpc_base.Request(
        'missing_event', params={'foo': 'bar'}, msg_id=1))
    assert response.error['code'] == -32601
    assert response.error['message'] == 'Method not found'

    response = server.receive_request(jsonrpc_base.Request(
        'on_server_event'))
    assert response is None

    def bad_handler():
        raise Exception("Bad Server Handler")
    server.on_bad_handler = bad_handler

    # receive_request will normally print traceback when an exception is caught
    # This isn't necessary for the test
    response = server.receive_request(jsonrpc_base.Request(
        'on_bad_handler', msg_id=1))
    assert response.error['code'] == -32000
    assert response.error['message'] == 'Server Error: Bad Server Handler'

    async def async_event_handler(*args, **kwargs):
        return args, kwargs
    server.on_async_server_event = async_event_handler

    response = server.receive_request(jsonrpc_base.Request(
        'on_async_server_event', msg_id=1))
    assert response.error['code'] == -32000
    assert response.error['message'] == (
        'Server Error: Async handlers are not'
        ' supported in synchronous sever implementations')


async def test_async_receive_server_requests(server):
    def event_handler(*args, **kwargs):
        return args, kwargs
    server.on_server_event = event_handler
    server.namespace.on_server_event = event_handler

    response = await server.async_receive_request(jsonrpc_base.Request(
        'on_server_event', msg_id=1))
    args, kwargs = response.result
    assert len(args) == 0
    assert len(kwargs) == 0

    async def async_event_handler(*args, **kwargs):
        return args, kwargs
    server.on_async_server_event = async_event_handler

    response = await server.async_receive_request(jsonrpc_base.Request(
        'on_async_server_event', msg_id=1))
    args, kwargs = response.result
    assert len(args) == 0
    assert len(kwargs) == 0

    response = await server.async_receive_request(jsonrpc_base.Request(
        'missing_event', params={'foo': 'bar'}, msg_id=1))
    assert response.error['code'] == -32601
    assert response.error['message'] == 'Method not found'

    response = await server.async_receive_request(jsonrpc_base.Request(
        'on_server_event'))
    assert response is None


async def test_server_responses(server):
    def handler(message):
        handler.response = message
    server._handler = handler

    def subtract(foo, bar):
        return foo - bar
    server.subtract = subtract

    response = server.receive_request(jsonrpc_base.Request(
        'subtract', params={'foo': 5, 'bar': 3}, msg_id=1))
    server.send_message(response)
    assertSameJSON(
        '''{"jsonrpc": "2.0", "result": 2, "id": 1}''',
        handler.response
    )

    response = server.receive_request(jsonrpc_base.Request(
        'subtract', params=[11, 7], msg_id=1))
    server.send_message(response)
    assertSameJSON(
        '''{"jsonrpc": "2.0", "result": 4, "id": 1}''',
        handler.response
    )

    response = server.receive_request(jsonrpc_base.Request(
        'missing_method', msg_id=1))
    server.send_message(response)
    assertSameJSON(
        '''{"jsonrpc": "2.0", "error": {"code": -32601, "message": "Method'''
        ''' not found"}, "id": 1}''',
        handler.response
    )

    def bad_handler(self):
        raise MockTransportError("Transport Error")
    server._handler = bad_handler

    def good_method():
        return True
    server.good_method = good_method
    response = server.receive_request(jsonrpc_base.Request(
        'good_method', msg_id=1))
    with pytest.raises(
            TransportError,
            match="Error responding to server method 'good_method': "
            "Transport Error"):
        server.send_message(response)

    async def async_bad_method():
        raise ValueError("Mock server error")
    server.async_bad_method = async_bad_method
    response = await server.async_receive_request(jsonrpc_base.Request(
        'async_bad_method', msg_id=1))
    with pytest.raises(
            TransportError,
            match="Error responding to server method 'async_bad_method': "
            "Transport Error"):
        server.send_message(response)


def test_base_message(server):
    message = jsonrpc_base.Message()
    assert message.response_id is None

    with pytest.raises(NotImplementedError):
        message.serialize()

    with pytest.raises(NotImplementedError):
        message.parse_response(None)

    with pytest.raises(NotImplementedError):
        message.transport_error_text

    with pytest.raises(NotImplementedError):
        str(message)


def test_request(server):
    with pytest.raises(
            ProtocolError,
            match='Request from server does not contain method'):
        jsonrpc_base.Request.parse({})

    with pytest.raises(
            ProtocolError,
            match='Parameters must either be a positional list or named dict.'
            ):
        jsonrpc_base.Request.parse(
            {'method': 'test_method', 'params': 'string_params'})

    request = jsonrpc_base.Request('test_method', msg_id=1)
    assert request.response_id == 1


def test_jsonrpc_1_0_call(server):
    # JSON-RPC 1.0 spec needs "error" to be present (with `null` value) when no
    # error occured
    def handler(message):
        assert message.params == [42, 23]
        return {
            "result": 19,
            "id": 1,
            "error": None,
        }

    server._handler = handler
    assert server.subtract(42, 23) == 19
